#define _CRT_SECURE_NO_WARNINGS
#include "Game.h"
#include "Sound.h"
#include "Math_Util.h"

#include <stdio.h>
#include <math.h>

#define HISCORE_FILENAME "hiscore.dat"

Game::Game()
	: m_is_active(false)
	, m_hiscore(0.f)
	, m_tunnel(new Tunnel())
	, m_tunnel_pos(0.f)
	, m_alpha_score(0.f)
	, m_alpha_menu(0.f)
{
	m_local_pos = D3DXVECTOR3(0.f, 0.f, 0.f);
	m_local_dir = D3DXVECTOR3(0.f, 0.f, 1.f);

	// Try to read the hiscore file
	FILE *f = fopen(HISCORE_FILENAME, "rb");
	if (f != 0)
	{
		float v1;
		if (fread(&v1, 1, sizeof(v1), f) == sizeof(float))
		{
			if (v1 > 0.f)
				m_hiscore = v1;
		}
		fclose(f);
	}
}

Game::~Game()
{
}

bool Game::Is_Running() const
{
	return m_is_active;
}

float Game::Get_Score() const
{
	return m_tunnel_pos * 0.01f;
}

void Game::Restart()
{
	m_is_active = true;
	m_tunnel_pos = 0.f;
	m_local_pos = D3DXVECTOR3(0.f, 0.f, 0.f);
	m_local_dir = D3DXVECTOR3(0.f, 0.f, 1.f);

	Sound::Get()->Play(Sound::Fx_Start);
}

void Game::Update(float delta_t, Input *input)
{
	if (m_is_active)
	{
		Update_Game(delta_t, input);
	}
	else
	{
		Update_Menu(delta_t, input);
	}
}

void Game::Render(Renderer *renderer)
{
	renderer->Set_Projection_Matrix();

	{
		D3DXMATRIX view_mtx;
		D3DXVECTOR3 tunnel_pos1 = m_tunnel->Get_Pos_At(m_tunnel_pos);
		D3DXVECTOR3 tunnel_pos2 = m_tunnel->Get_Pos_At(m_tunnel_pos + 0.1f);
//		D3DXVECTOR3 pos2 = pos1 + D3DXVECTOR3(1, 0, 1);
//		D3DXVECTOR3 pos2 = pos1 + D3DXVECTOR3(0, 0, 1);
//		pos1 += D3DXVECTOR3(0, 15, 0);

		D3DXMATRIX tunnel_mtx;
		D3DXVECTOR3 up(0.f, 1.f, 0.f);
		D3DXMatrixLookAtLH(&tunnel_mtx, &tunnel_pos1, &tunnel_pos2, &up);

		D3DXMATRIX local_mtx;
		D3DXVECTOR3 local_dest = m_local_pos + m_local_dir;
		D3DXMatrixLookAtLH(&local_mtx, &m_local_pos, &local_dest, &up);

		renderer->Set_View_Matrix(tunnel_mtx * local_mtx);
	}

	D3DXMATRIX world_mtx;
	D3DXMatrixIdentity(&world_mtx);
	renderer->Set_World_Matrix(world_mtx);

	// Calculate the fog density
	{
		float fog_exp = 0.4f - m_tunnel_pos * 0.001f;
		renderer->Set_Param("fog_exp", Clamp(fog_exp, 0.1f, 0.4f));
	}

	// Render the tunnel
	renderer->Render_Mesh(m_tunnel->Get_Tunnel_Mesh());

	// Write the score
	{
		WCHAR str[256];
		int dist = (int)(Get_Score() * 100.f);

		wsprintf(str, L"Intestinal cavity explored: %d.%02d meters", dist / 100, dist % 100);
		renderer->Write_Text(10, 10, str, D3DCOLOR_ARGB((int)(m_alpha_score * 255.f), 246, 234, 120), false, false);
	}

	// Write the menu
	if (m_alpha_menu > 0.001f)
	{
		D3DCOLOR col = D3DCOLOR_ARGB((int)(m_alpha_menu * 255.f), 246, 234, 120);
		int center_x = renderer->Get_Width() / 2;
		int h = 20;
		int y;

		renderer->Write_Text(center_x, 30, L"Intestinal Cavity Explorer", col, true, true);

		y = (int)(renderer->Get_Height() * 0.3f);
		renderer->Write_Text(center_x, y, L"Left mouse click to start the game", col, true, false); y += h;
		renderer->Write_Text(center_x, y, L"Right mouse click to quit", col, true, false); y += h;
		y += h;
		renderer->Write_Text(center_x, y, L"Move the mouse to explore and to avoid the walls =)", col, true, false); y += h;

		{ // Write the hiscore
			WCHAR str[256];
			int dist = (int)(m_hiscore * 100.f);

			wsprintf(str, L"Farthest intestination: %d.%02d meters", dist / 100, dist % 100);
			y = (int)(renderer->Get_Height() * 0.8f);
			renderer->Write_Text(center_x, y, str, col, true, false);
		}

		col = D3DCOLOR_ARGB((int)(m_alpha_menu * 0.3f * 255.f), 246, 234, 120);
		renderer->Write_Text(center_x, renderer->Get_Height() - h - h, L"LD48h #16 entry, 2009, Jere \"XMunkki\" Sanisalo --  http://www.xmunkki.org/", col, true, false); y += h;
	}
}

void Game::Update_Menu(float delta_t, Input *input)
{
	// Update the UI
	if (Get_Score() > 0.f)
		m_alpha_score = Lerp(m_alpha_score, 0.f, 0.01f);
	m_alpha_menu = Lerp(m_alpha_menu, 1.f, 0.1f);

	// Restart the level?
	if (input->Left_Mouse_Pressed())
		Restart();

	// Quit the game?
	if (input->Right_Mouse_Pressed())
		PostQuitMessage(0);
}

void Game::Update_Game(float delta_t, Input *input)
{
	// Update the UI
	m_alpha_score = Lerp(m_alpha_score, 1.f, 0.1f);
	m_alpha_menu = Lerp(m_alpha_menu, 0.f, 0.07f);

	// Calculate the movement speed
	float movement_speed = 22.5f + powf(m_tunnel_pos, 0.5f);
	movement_speed = Clamp(movement_speed, 0.f, 45.f);

	// User control for the direction
	{
		D3DXMATRIX mtx;
		D3DXVECTOR2 mouse_ctrl = input->Get_Mouse_Delta() * 0.003f;
//		D3DXVECTOR2 mouse_ctrl = D3DXVECTOR2(0.f, 0.1f);
		D3DXMatrixRotationYawPitchRoll(&mtx, mouse_ctrl.x, mouse_ctrl.y, 0.f);

		D3DXVECTOR4 new_dir;
		D3DXVec3Transform(&new_dir, &m_local_dir, &mtx);
		m_local_dir = Vec4_To_Vec3(new_dir);
		D3DXVec3Normalize(&m_local_dir, &m_local_dir);
	}

	// Force the local dir to watch mostly ahead
	while (m_local_dir.z < 0.9f)
	{
		m_local_dir.z = 0.9f;
		D3DXVec3Normalize(&m_local_dir, &m_local_dir);
	}

	// Move the local position (keep it in the xy-plane)
	m_local_pos += m_local_dir * (movement_speed * delta_t);
	float dist_moved_fwd = m_local_pos.z;
	m_local_pos.z = 0.f;

	// Advance the tunnel position (iteratively search the correct amount to move)
	D3DXMATRIX frame_prev;
	m_tunnel->Get_Frame_At(frame_prev, m_tunnel_pos);

	{
		float left = dist_moved_fwd;

		while (left > 0.001f)
		{
			D3DXVECTOR3 pos1 = m_tunnel->Get_Pos_At(m_tunnel_pos);
			float amt = 0.1f;
			D3DXVECTOR3 pos2 = m_tunnel->Get_Pos_At(m_tunnel_pos + amt);

			float moved = Vec_Dist(pos1, pos2);
			if (moved <= left)
			{
				m_tunnel_pos += amt;
				left -= moved;
			}
			else
			{
				m_tunnel_pos += amt * (left / moved);
				left = 0.f;
			}
		}
	}

	D3DXMATRIX frame_cur;
	m_tunnel->Get_Frame_At(frame_cur, m_tunnel_pos);

	// Turn the local viewing direction according to how much the tunnel turned
	{
		// Calculate the rotation from 'frame_prev' to 'frame_cur'
		D3DXMATRIX inv_prev;
		D3DXMatrixInverse(&inv_prev, NULL, &frame_prev);

		D3DXMATRIX rot_mtx;
		D3DXMatrixMultiply(&rot_mtx, &inv_prev, &frame_cur);

		// Rotate the looking direction according to the change in the angles
		D3DXVECTOR4 dir = Vec3_To_Vec4(m_local_dir, 0.f);
		D3DXVECTOR4 new_dir;
		D3DXVec4Transform(&new_dir, &dir, &rot_mtx);

		m_local_dir = Vec4_To_Vec3(new_dir);
	}

	// Proximity test
	{
		float prox_radius = m_tunnel->Get_Radius_At(m_tunnel_pos);
		float dist_from_center = D3DXVec3Length(&m_local_pos);
		if ((prox_radius - dist_from_center) < 0.1f)
		{ // Collision!
			// Fix back into the field
			dist_from_center = prox_radius - 0.1f;
			if (dist_from_center > 0.f)
			{
				D3DXVec3Normalize(&m_local_pos, &m_local_pos);
				m_local_pos *= dist_from_center;
			}
			else
			{
				m_local_pos = D3DXVECTOR3(0.f, 0.f, 0.f);
			}

			// Play the sound
			Sound::Get()->Play(Sound::Fx_Die);

			// Made a new hiscore?
			if (Get_Score() > m_hiscore)
			{
				m_hiscore = Get_Score();

				// Save the score
				FILE *f = fopen(HISCORE_FILENAME, "wb");
				if (f != 0)
				{
					fwrite(&m_hiscore, 1, sizeof(m_hiscore), f);
					fclose(f);
				}
			}

			// No longer running
			m_is_active = false;
		}
	}

	// Update the tunnel rendering
	m_tunnel->Update_Mesh(m_tunnel_pos);
}
